"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = require("events");
const IORedis = require("ioredis");
const os = require("os");
const uniqid = require("uniqid");
const url = require("url");
const E_ACCEPT_TIMEOUT = new Error('Accept timeout');
const E_CALL_TIMEOUT = new Error('Call timeout');
const parserRedisURI = (uri) => {
    const parsedURI = url.parse(uri, true);
    return Object.assign({ host: parsedURI.hostname, port: parseInt(parsedURI.port || '6379', 10), password: parsedURI.auth }, parsedURI.query);
};
class Pool extends events_1.EventEmitter {
    constructor({ channelPrefix = 'swimer', Swimer, identity }, redisOptionsOrURI) {
        super();
        this._contexts = {};
        this._swimers = {};
        this._acquireQueue = {};
        this._connected = false;
        this._channelPrefix = channelPrefix;
        this._identity = identity;
        this.id = uniqid.process(`poolid-${channelPrefix}-`);
        if (typeof redisOptionsOrURI === 'string') {
            redisOptionsOrURI = parserRedisURI(redisOptionsOrURI);
        }
        this._subscriber = new IORedis(redisOptionsOrURI);
        this._publisher = new IORedis(redisOptionsOrURI);
        this._subscribe();
        this._Swimer = Swimer;
        const pool = this;
        class ShadowSwimer {
            constructor(id) {
                this._id = id;
                return new Proxy(this, {
                    get(target, key, receiver) {
                        const prop = Reflect.get(target, key, receiver);
                        if (prop || ['then', 'inspect'].includes(key) || typeof key === 'symbol') {
                            return prop;
                        }
                        return async (...args) => {
                            return await pool._ask('call', {
                                sid: target._id,
                                method: key,
                                args,
                            });
                        };
                    },
                });
            }
        }
        this._ShadowSwimer = ShadowSwimer;
        this._acquireQueueInterval = setInterval(() => {
            if (!this._connected) {
                return;
            }
            const ids = Object.keys(this._acquireQueue);
            const contexts = Object.keys(this._contexts);
            if (this._destroyed && ids.length === 0 && contexts.length === 0) {
                this._destroy();
                return;
            }
            ids.forEach(id => {
                if (this._acquireQueue[id].length > 0) {
                    if (this._acquireQueue[id][0].pending) {
                        return;
                    }
                    const { reject, resolve, args, shadow, remote } = this._acquireQueue[id][0];
                    this._acquireQueue[id][0].pending = true;
                    if (remote) {
                        if (this._swimers[id]) {
                            resolve();
                        }
                        else {
                            reject();
                        }
                        this._acquireQueue[id].shift();
                    }
                    else {
                        this._getSwimer(id, args, shadow).then(swimer => {
                            resolve(swimer);
                            this._acquireQueue[id].shift();
                        }).catch(error => {
                            reject(error);
                            this._acquireQueue[id].shift();
                        });
                    }
                }
                else {
                    Reflect.deleteProperty(this._acquireQueue, id);
                }
            });
        });
        this._scavengerInterval = setInterval(() => {
            const deadline = Date.now() - 600000;
            Object.keys(this._swimers).forEach(sid => {
                const swimerWrapper = this._swimers[sid];
                if (swimerWrapper.activatedAt <= deadline) {
                    this._removeSwimer(sid);
                }
            });
        }, 300000);
    }
    async getSwimer(args, shadow = false) {
        const id = this._identity(args);
        return new Promise(((resolve, reject) => {
            if (!this._acquireQueue[id]) {
                this._acquireQueue[id] = [];
            }
            this._acquireQueue[id].push({ resolve, reject, args, shadow });
        }));
    }
    async forceGetSwimer(args) {
        const id = this._identity(args);
        if (this._swimers[id]) {
            this._swimers[id].activatedAt = Date.now();
            return this._swimers[id].swimer;
        }
        await this.destruct(args);
        return this._newSwimer(id, args);
    }
    async whereis(args) {
        return await this._topic('whereis', { sid: this._identity(args) });
    }
    async destruct(args) {
        const sid = this._identity(args);
        if (this._swimers[sid]) {
            this._removeSwimer(sid);
        }
        else {
            await this._topic('destruct', { sid });
        }
    }
    async state(pid) {
        if (pid) {
            return await this._ask('state', { pid });
        }
        else {
            return await this._topic('state');
        }
    }
    destroy() {
        this._destroyed = true;
    }
    _destroy() {
        clearInterval(this._acquireQueueInterval);
        clearInterval(this._scavengerInterval);
        this._subscriber.disconnect();
        this._publisher.disconnect();
    }
    async _getSwimer(id, args, shadow) {
        if (this._swimers[id]) {
            this._swimers[id].activatedAt = Date.now();
            return this._swimers[id].swimer;
        }
        try {
            await this._ask('born', { sid: id });
            return new this._ShadowSwimer(id);
        }
        catch (err) {
            if (err === E_ACCEPT_TIMEOUT && !shadow) {
                return this._newSwimer(id, args);
            }
            else {
                throw err;
            }
        }
    }
    _newSwimer(id, args) {
        const swimer = new this._Swimer(args);
        this._swimers[id] = {
            swimer,
            activatedAt: Date.now(),
            createdAt: Date.now(),
        };
        return swimer;
    }
    _processQuestion({ question, params = {}, qid, pid }) {
        const sid = params.sid;
        const swimerWrapper = this._swimers[sid];
        switch (question) {
            case 'whereis':
                if (swimerWrapper) {
                    this._reply(pid, qid, 'answer', `${sid} is here`);
                }
                break;
            case 'state':
                if (params.pid) {
                    if (this.id === params.pid) {
                        this._reply(pid, qid, 'accept');
                        this._reply(pid, qid, 'answer', Object.keys(this._swimers).map(swimerId => {
                            return {
                                sid: swimerId,
                                activatedAt: this._swimers[swimerId].activatedAt,
                                createdAt: this._swimers[swimerId].createdAt,
                            };
                        }));
                    }
                }
                else {
                    this._reply(pid, qid, 'answer', {
                        count: Object.keys(this._swimers).length,
                        hostname: os.hostname(),
                    });
                }
                break;
            case 'destruct':
                if (swimerWrapper) {
                    this._removeSwimer(sid);
                    this._reply(pid, qid, 'answer', { sid });
                }
                break;
            case 'born':
            case 'call':
                if (swimerWrapper) {
                    swimerWrapper.activatedAt = Date.now();
                }
                if (pid === this.id) {
                    return;
                }
                const acquireQueue = this._acquireQueue[sid];
                if (swimerWrapper) {
                    const swimer = swimerWrapper.swimer;
                    this._reply(pid, qid, 'accept', { sid });
                    switch (question) {
                        case 'born':
                            this._reply(pid, qid, 'answer', { sid });
                            break;
                        case 'call':
                            const method = swimer[params.method];
                            if (method) {
                                (async () => {
                                    const result = await method.bind(swimer)(...params.args);
                                    this._reply(pid, qid, 'answer', result);
                                })().catch(error => {
                                    this._reply(pid, qid, 'reject', {
                                        name: error.name,
                                        message: error.message,
                                        stack: error.stack,
                                    });
                                });
                            }
                            else {
                                const error = new Error(`Method [${params.method}] is not exists.`);
                                this._reply(pid, qid, 'reject', {
                                    name: error.name,
                                    message: error.message,
                                    stack: error.stack,
                                });
                            }
                            break;
                        default:
                    }
                }
                else if (question === 'born' && acquireQueue && acquireQueue.length > 0) {
                    if (this.id > pid) {
                        this._reply(pid, qid, 'accept', { sid });
                    }
                    acquireQueue.push({
                        resolve: () => {
                            this._reply(pid, qid, 'answer', { sid });
                        },
                        reject: () => {
                        },
                        remote: true,
                    });
                }
                break;
            default:
        }
    }
    _removeSwimer(sid) {
        const swimerWrapper = this._swimers[sid];
        if (swimerWrapper) {
            Reflect.deleteProperty(this._swimers, sid);
            const swimer = swimerWrapper.swimer;
            if (swimer.destroy) {
                swimer.destroy();
            }
        }
    }
    _processReply({ reply = {}, qid, type, pid }) {
        if (type === 'excess') {
            this._removeSwimer(reply.sid);
        }
        else if (this._contexts[qid]) {
            const context = this._contexts[qid];
            switch (context.type) {
                case 'ask':
                    switch (type) {
                        case 'accept':
                            clearTimeout(context.acceptTimer);
                            break;
                        case 'answer':
                            clearTimeout(context.answerTimer);
                            context.resolve(reply);
                            Reflect.deleteProperty(this._contexts, qid);
                            break;
                        case 'reject':
                            clearTimeout(context.answerTimer);
                            const error = new Error(reply.message);
                            error.name = reply.name;
                            error.stack = reply.stack;
                            context.reject(error);
                            Reflect.deleteProperty(this._contexts, qid);
                            break;
                        default:
                    }
                    break;
                case 'topic':
                    switch (type) {
                        case 'answer':
                            context.chunks[pid] = reply;
                            break;
                        default:
                    }
            }
        }
        else if (['answer', 'reject'].includes(type)) {
            this._reply(this.id, qid, 'excess', reply);
        }
    }
    _subscribe() {
        this._subscriber.subscribe(`${this._channelPrefix}.${this.id}`, `${this._channelPrefix}.question`, () => {
            setTimeout(() => {
                this.emit('connected', this.id);
                this._connected = true;
            }, 750);
        });
        this._subscriber.on('message', (topic, data) => {
            try {
                const jsonData = JSON.parse(data);
                switch (topic) {
                    case `${this._channelPrefix}.question`:
                        this._processQuestion(jsonData);
                        break;
                    case `${this._channelPrefix}.${this.id}`:
                        this._processReply(jsonData);
                        break;
                    default:
                }
            }
            catch (err) {
                err.message = `${topic}. Parse data failed. ${err.message}`;
                this.emit('error', err);
            }
        });
    }
    async _ask(question, params) {
        return await new Promise(((resolve, reject) => {
            const qid = uniqid.process('question-');
            const answerTimer = setTimeout(() => {
                Reflect.deleteProperty(this._contexts, qid);
                reject(E_CALL_TIMEOUT);
            }, 30000);
            const acceptTimer = setTimeout(() => {
                Reflect.deleteProperty(this._contexts, qid);
                reject(E_ACCEPT_TIMEOUT);
                clearTimeout(answerTimer);
            }, 500);
            this._contexts[qid] = {
                type: 'ask',
                resolve,
                reject,
                acceptTimer,
                answerTimer,
            };
            this._publisher.publish(`${this._channelPrefix}.question`, JSON.stringify({
                question,
                params,
                qid,
                pid: this.id,
            }));
        }));
    }
    async _topic(question, params, waitFor = 500) {
        return await new Promise(((resolve, reject) => {
            const qid = uniqid.process('question-');
            const chunks = {};
            const finishTimer = setTimeout(() => {
                Reflect.deleteProperty(this._contexts, qid);
                resolve(Object.keys(chunks).map(pid => {
                    return {
                        pid,
                        reply: chunks[pid],
                    };
                }));
            }, waitFor);
            this._contexts[qid] = {
                type: 'topic',
                resolve,
                reject,
                chunks,
                finishTimer,
            };
            this._publisher.publish(`${this._channelPrefix}.question`, JSON.stringify({
                question,
                params,
                qid,
                pid: this.id,
            }));
        }));
    }
    _reply(pid, qid, type, data) {
        this._publisher.publish(`${this._channelPrefix}.${pid}`, JSON.stringify({
            pid: this.id,
            qid,
            reply: data,
            type,
        }));
    }
}
exports.default = Pool;
//# sourceMappingURL=index.js.map