import { EventEmitter } from 'events';
import { Redis, RedisOptions } from 'ioredis';
import * as IORedis from 'ioredis';
import * as os from 'os';
import * as uniqid from 'uniqid';
import * as url from 'url';
import Timer = NodeJS.Timer;

const E_ACCEPT_TIMEOUT = new Error('Accept timeout');
const E_CALL_TIMEOUT = new Error('Call timeout');

export interface ISwimer {
  [property: string]: any;
}

export interface ISwimerConstructor<Args, T extends ISwimer> {
  new(args: Args): T;
}

export interface IPoolOptions<Args, T extends ISwimer> {
  channelPrefix?: string;
  Swimer: ISwimerConstructor<Args, T>;
  identity: (args: Args) => string;
}

interface IObject {
  [key: string]: any;
}

interface IQuestionContext {
  question: 'born' | 'call' | 'whereis' | 'state' | 'destruct';
  params: IObject;
  qid: string;
  pid: string;
}

interface IReplyContext {
  pid: string;
  qid: string;
  type: 'accept' | 'answer' | 'reject' | 'excess';
  reply?: any;
}

interface IContext {
  resolve: (value?: any) => void;
  reject: (reason?: any) => void;
  type: 'ask' | 'topic';
  acceptTimer?: Timer;
  answerTimer?: Timer;
  chunks?: IObject;
  finishTimer?: Timer;
}

interface IAcquireContext<Args> {
  resolve: (value?: any) => void;
  reject: (reason?: any) => void;
  args?: Args;
  shadow?: boolean;
  pending?: boolean;
  remote?: boolean;
}

const parserRedisURI = (uri: string): RedisOptions => {
  const parsedURI = url.parse(uri, true);
  return {
    host: parsedURI.hostname,
    port: parseInt(parsedURI.port || '6379', 10),
    password: parsedURI.auth,
    ...parsedURI.query,
  };
};

export default class Pool<Args extends any, T extends ISwimer> extends EventEmitter {
  public id: string;
  private _subscriber: Redis;
  private _publisher: Redis;
  private _contexts: {
    [id: string]: IContext;
  } = {};
  private _swimers: {
    [id: string]: {
      swimer: T;
      activatedAt: number;
      createdAt: number;
    };
  } = {};
  private _channelPrefix: string;
  private _Swimer: ISwimerConstructor<Args, T>;
  private _ShadowSwimer: { new(id: string): any };
  private _acquireQueue: {
    [id: string]: Array<IAcquireContext<Args>>;
  } = {};
  private _acquireQueueInterval: number;
  private _destroyed?: boolean;
  private _identity: (args: Args) => string;
  private _scavengerInterval: Timer;
  private _connected: boolean = false;

  constructor({ channelPrefix = 'swimer', Swimer, identity }: IPoolOptions<Args, T>, redisOptionsOrURI?: RedisOptions | string) {
    super();
    this._channelPrefix = channelPrefix;
    this._identity = identity;
    this.id = uniqid.process(`poolid-${channelPrefix}-`);

    if (typeof redisOptionsOrURI === 'string') {
      redisOptionsOrURI = parserRedisURI(redisOptionsOrURI);
    }

    this._subscriber = new IORedis(redisOptionsOrURI);
    this._publisher = new IORedis(redisOptionsOrURI);

    this._subscribe();

    this._Swimer = Swimer;

    const pool = this;

    class ShadowSwimer {
      public _id: string;

      constructor(id: string) {
        this._id = id;
        return new Proxy(this, {
          get(target, key: string, receiver) {
            const prop = Reflect.get(target, key, receiver);
            if (prop || ['then', 'inspect'].includes(key) || typeof key === 'symbol') {
              return prop;
            }
            return async (...args: any[]) => {
              return await pool._ask('call', {
                sid: target._id,
                method: key,
                args,
              });
            };
          },
        });
      }
    }

    this._ShadowSwimer = ShadowSwimer;

    this._acquireQueueInterval = setInterval(() => {
      if (!this._connected) {
        return;
      }
      const ids = Object.keys(this._acquireQueue);
      const contexts = Object.keys(this._contexts);
      if (this._destroyed && ids.length === 0 && contexts.length === 0) {
        this._destroy();
        return;
      }
      ids.forEach(id => {
        if (this._acquireQueue[id].length > 0) {
          if (this._acquireQueue[id][0].pending) {
            return;
          }
          const { reject, resolve, args, shadow, remote } = this._acquireQueue[id][0];
          this._acquireQueue[id][0].pending = true;
          if (remote) {
            if (this._swimers[id]) {
              resolve();
            } else {
              reject();
            }
            this._acquireQueue[id].shift();
          } else {
            this._getSwimer(id, args!, shadow!).then(swimer => {
              resolve(swimer);
              this._acquireQueue[id].shift();
            }).catch(error => {
              reject(error);
              this._acquireQueue[id].shift();
            });
          }
        } else {
          Reflect.deleteProperty(this._acquireQueue, id);
        }
      });
    });

    this._scavengerInterval = setInterval(() => {
      const deadline = Date.now() - 600000;
      Object.keys(this._swimers).forEach(sid => {
        const swimerWrapper = this._swimers[sid];
        if (swimerWrapper.activatedAt <= deadline) {
          this._removeSwimer(sid);
        }
      });
    }, 300000);
  }

  public async getSwimer(args: Args, shadow: boolean = false): Promise<T> {
    const id = this._identity(args);
    return new Promise<T>(((resolve, reject) => {
      if (!this._acquireQueue[id]) {
        this._acquireQueue[id] = [];
      }
      this._acquireQueue[id].push({ resolve, reject, args, shadow });
    }));
  }

  public async forceGetSwimer(args: Args): Promise<T> {
    const id = this._identity(args);
    if (this._swimers[id]) {
      this._swimers[id].activatedAt = Date.now();
      return this._swimers[id].swimer;
    }
    await this.destruct(args);
    return this._newSwimer(id, args);
  }

  public async whereis(args: Args): Promise<Array<{ pid: string, reply: string }>> {
    return await this._topic('whereis', { sid: this._identity(args) });
  }

  public async destruct(args: Args) {
    const sid = this._identity(args);
    if (this._swimers[sid]) {
      this._removeSwimer(sid);
    } else {
      await this._topic('destruct', { sid });
    }
  }

  public state(pid: string): Promise<Array<{
    sid: string;
    activatedAt: number;
    createdAt: number;
  }>>;
  public state(): Promise<Array<{
    pid: string;
    reply: {
      hostname: string;
      count: number;
    }
  }>>;
  public async state(pid?: string): Promise<any> {
    if (pid) {
      return await this._ask('state', { pid });
    } else {
      return await this._topic('state');
    }
  }

  public destroy() {
    this._destroyed = true;
  }

  private _destroy() {
    clearInterval(this._acquireQueueInterval);
    clearInterval(this._scavengerInterval);
    this._subscriber.disconnect();
    this._publisher.disconnect();
  }

  private async _getSwimer(id: string, args: Args, shadow: boolean): Promise<T> {
    if (this._swimers[id]) {
      this._swimers[id].activatedAt = Date.now();
      return this._swimers[id].swimer;
    }
    try {
      await this._ask('born', { sid: id });
      return new this._ShadowSwimer(id);
    } catch (err) {
      if (err === E_ACCEPT_TIMEOUT && !shadow) {
        return this._newSwimer(id, args);
      } else {
        throw err;
      }
    }
  }

  private _newSwimer(id: string, args: Args): T {
    const swimer = new this._Swimer(args);
    this._swimers[id] = {
      swimer,
      activatedAt: Date.now(),
      createdAt: Date.now(),
    };
    return swimer;
  }

  private _processQuestion({ question, params = {}, qid, pid }: IQuestionContext) {
    const sid = params.sid;
    const swimerWrapper = this._swimers[sid];
    switch (question) {
      case 'whereis':
        if (swimerWrapper) {
          this._reply(pid, qid, 'answer', `${sid} is here`);
        }
        break;
      case 'state':
        if (params.pid) {
          if (this.id === params.pid) {
            this._reply(pid, qid, 'accept');
            this._reply(pid, qid, 'answer', Object.keys(this._swimers).map(swimerId => {
              return {
                sid: swimerId,
                activatedAt: this._swimers[swimerId].activatedAt,
                createdAt: this._swimers[swimerId].createdAt,
              };
            }));
          }
        } else {
          this._reply(pid, qid, 'answer', {
            count: Object.keys(this._swimers).length,
            hostname: os.hostname(),
          });
        }
        break;
      case 'destruct':
        if (swimerWrapper) {
          this._removeSwimer(sid);
          this._reply(pid, qid, 'answer', { sid });
        }
        break;
      case 'born':
      case 'call':
        if (swimerWrapper) {
          swimerWrapper.activatedAt = Date.now();
        }
        if (pid === this.id) {
          return;
        }
        const acquireQueue = this._acquireQueue[sid];
        if (swimerWrapper) {
          const swimer = swimerWrapper.swimer;
          this._reply(pid, qid, 'accept', { sid });
          switch (question) {
            case 'born':
              // 告诉提问者，该对象已经存在
              this._reply(pid, qid, 'answer', { sid });
              break;
            case 'call':
              // 调用swimer的方法
              const method = swimer[params.method];
              if (method) {
                (async () => {
                  const result = await method.bind(swimer)(...params.args);
                  this._reply(pid, qid, 'answer', result);
                })().catch(error => {
                  // 执行方法报错
                  this._reply(pid, qid, 'reject', {
                    name: error.name,
                    message: error.message,
                    stack: error.stack,
                  });
                });
              } else {
                // 方法不存在，报错
                const error = new Error(`Method [${params.method}] is not exists.`);
                this._reply(pid, qid, 'reject', {
                  name: error.name,
                  message: error.message,
                  stack: error.stack,
                });
              }
              break;
            default:
          }
        } else if (question === 'born' && acquireQueue && acquireQueue.length > 0) {
          if (this.id > pid) {
            this._reply(pid, qid, 'accept', { sid });
          }
          acquireQueue.push({
            resolve: () => {
              this._reply(pid, qid, 'answer', { sid });
            },
            reject: () => {
              // noop.
            },
            remote: true,
          });
        }
        break;
      default:
    }
  }

  private _removeSwimer(sid: string) {
    const swimerWrapper = this._swimers[sid];
    if (swimerWrapper) {
      Reflect.deleteProperty(this._swimers, sid);
      const swimer = swimerWrapper.swimer;
      if (swimer.destroy) {
        swimer.destroy();
      }
    }
  }

  private _processReply({ reply = {}, qid, type, pid }: IReplyContext) {
    if (type === 'excess') {
      // 如果是多余的对象，则销毁掉
      this._removeSwimer(reply.sid);
    } else if (this._contexts[qid]) {
      const context = this._contexts[qid];
      switch (context.type) {
        case 'ask':
          switch (type) {
            case 'accept':
              clearTimeout(context.acceptTimer!);
              break;
            case 'answer':
              clearTimeout(context.answerTimer!);
              context.resolve(reply);
              Reflect.deleteProperty(this._contexts, qid);
              break;
            case 'reject':
              clearTimeout(context.answerTimer!);
              const error = new Error(reply.message);
              error.name = reply.name;
              error.stack = reply.stack;
              context.reject(error);
              Reflect.deleteProperty(this._contexts, qid);
              break;
            default:
          }
          break;
        case 'topic':
          switch (type) {
            case 'answer':
              context.chunks![pid] = reply;
              break;
            default:
          }
      }
    } else if (['answer', 'reject'].includes(type)) {
      // 如果找不到上下文，说明上下文已经被执行了，那么这一次答复是多余的
      this._reply(this.id, qid, 'excess', reply);
    }
  }

  private _subscribe() {
    this._subscriber.subscribe(`${this._channelPrefix}.${this.id}`, `${this._channelPrefix}.question`, () => {
      // 订阅成功后，等待750ms再开始向外发送询问，以保证不会因为其他节点没接收到订阅成功前的询问，而导致创建出多余的对象的问题
      setTimeout(() => {
        this.emit('connected', this.id);
        this._connected = true;
      }, 750);
    });

    this._subscriber.on('message', (topic, data) => {
      try {
        const jsonData = JSON.parse(data);
        switch (topic) {
          case `${this._channelPrefix}.question`:
            this._processQuestion(jsonData as IQuestionContext);
            break;
          case `${this._channelPrefix}.${this.id}`:
            this._processReply(jsonData as IReplyContext);
            break;
          default:
        }
      } catch (err) {
        err.message = `${topic}. Parse data failed. ${err.message}`;
        this.emit('error', err);
      }
    });
  }

  private async _ask(question: string, params?: IObject) {
    return await new Promise<any>(((resolve, reject) => {
      const qid = uniqid.process('question-');
      const answerTimer = setTimeout(() => {
        Reflect.deleteProperty(this._contexts, qid);
        reject(E_CALL_TIMEOUT);
      }, 30000);
      const acceptTimer = setTimeout(() => {
        Reflect.deleteProperty(this._contexts, qid);
        reject(E_ACCEPT_TIMEOUT);
        clearTimeout(answerTimer);
      }, 500);
      this._contexts[qid] = {
        type: 'ask',
        resolve,
        reject,
        acceptTimer,
        answerTimer,
      };
      this._publisher.publish(`${this._channelPrefix}.question`, JSON.stringify({
        question,
        params,
        qid,
        pid: this.id,
      } as IQuestionContext));
    }));
  }

  private async _topic(question: string, params?: IObject, waitFor: number = 500) {
    return await new Promise<any>(((resolve, reject) => {
      const qid = uniqid.process('question-');
      const chunks: IObject = {};
      const finishTimer = setTimeout(() => {
        Reflect.deleteProperty(this._contexts, qid);
        resolve(Object.keys(chunks).map(pid => {
          return {
            pid,
            reply: chunks[pid],
          };
        }));
      }, waitFor);
      this._contexts[qid] = {
        type: 'topic',
        resolve,
        reject,
        chunks,
        finishTimer,
      };
      this._publisher.publish(`${this._channelPrefix}.question`, JSON.stringify({
        question,
        params,
        qid,
        pid: this.id,
      } as IQuestionContext));
    }));
  }

  private _reply(pid: string, qid: string, type: 'answer' | 'accept' | 'reject' | 'excess', data?: any) {
    this._publisher.publish(`${this._channelPrefix}.${pid}`, JSON.stringify({
      pid: this.id,
      qid,
      reply: data,
      type,
    } as IReplyContext));
  }
}
